# *
# *	Copyright (c) 2000 Alberto Reggiori / <alberto.reggiori@jrc.it>
# *	ISIS/RIT, Joint Research Center Ispra (I)
# *
# * NOTICE
# *
# * This product is distributed under a BSD/ASF like license as described in the 'LICENSE'
# * file you should have received together with this source code. If you did not get a
# * a copy of such a license agreement you can pick up one at:
# *
# *     http://xml.jrc.it/RDFStore/LICENSE
# *
# *
package RDF::Store::Model;
{
use Carp;
use RDF::Digest;
use RDF::Model;
use RDF::Store::Resource;
use RDF::Store::Literal;
use RDF::Store::Statement;
use RDF::Store::NodeFactory;
use RDF::Store::SetModel;
use RDF::Digest::Util;
use RDF::Store::FindIndex;

@RDF::Store::Model::ISA = qw( RDF::Store::Resource RDF::Model RDF::Digest );

sub new {
	my ($pkg,$factory_or_uri,$triples,$findIndex,$shared) = @_;

    	my $self = $pkg->SUPER::new();

    	#whether triples and lookup are shared
    	$self->{shared}=(defined $shared) ? $shared : 0;
    	$self->{uri}=$factory_or_uri
		if(	(defined $factory_or_uri) &&
                        (!($factory_or_uri->isa("RDF::NodeFactory"))) );
    	# first find operation creates lookup table
    	$self->{nodeFactory}=(	(defined $factory_or_uri) && 
				($factory_or_uri->isa("RDF::NodeFactory"))) ? 
				$factory_or_uri : new RDF::Store::NodeFactory();

	# creates lookup table
	if( (!(defined $triples)) && (!(defined $findIndex)) ) {
		$self->{triples}={};
		$self->{_findIndex} = new RDF::Store::FindIndex();
	} else {
		$self->{triples}=$triples;
		$self->{_findIndex} = ( 	(defined $findIndex) && ($findIndex->isa("RDF::Store::FindIndex")) ) ?
						$findIndex : 
						undef;
	};

    	bless $self,$pkg;
};

# Set a base URI for the message.
# Affects creating of new resources and serialization syntax.
# <code>null</code> is just alright, meaning that
# the serializer will use always rdf:about and never rdf:ID
sub setSourceURI {
	$_[0]->{uri}=$_[1];
};

# Returns current base URI setting
sub getSourceURI {
	return $_[0]->{uri};
};

# Model access
#
# Number of triples in the model
# return  number of triples
sub size {
	# really not efficient :-(
	# we should use BerkeleyDB or DB_File tricks hereish...
	return scalar keys %{$_[0]->{triples}};
};

#watch out in sub implementations with DBMS(3)!!!!!!!!
sub isEmpty {
	# not efficient also - see size()
	return ($_[0]->size() > 0) ? 0 : 1;
};

sub elements {
	return $_[0]->{triples};
};

# Tests if the model contains the given triple.
# return  true if the triple belongs to the model;
# false otherwise.
# !!!watch out in sub implementations with DBMS(3)!!!!!!!!
sub contains {
	if($_[1]->isa("RDF::Statement")) {
		#index by label then ;-)
		return (exists $_[0]->{triples}->{$_[1]->hashCode()}) ? 1 : 0;
	} else {
		return (exists $_[0]->{triples}->{$_[1]}) ? 1 : 0;
	};
};

# Model manipulation: add, remove, find
#
# Adds a new triple to the model
sub add {
	croak "Subject or Statement ".$_[1]." is not either instance of RDF::Statement or RDF::Resource"
                unless( (defined $_[1]) && (($_[1]->isa('RDF::Resource')) || ($_[1]->isa('RDF::Statement'))) );
	croak "Predicate ".$_[2]." is not instance of RDF::Resource"
                unless( (not(defined $_[2])) || ($_[2]->isa('RDF::Resource')) );

	my $t;
	if($_[1]->isa("RDF::Statement")) {
		$t = $_[1];
	} elsif($_[1]->isa("RDF::Resource")) {
		if($_[3]->isa("RDF::RDFNode")) {
			$t = $_[0]->{nodeFactory}->createStatement($_[1],$_[2],$_[3]);
		} else { #otherwise create a literal
			$t = $_[0]->{nodeFactory}->createStatement($_[1],$_[2],
						$_[0]->{nodeFactory}->createLiteral($_[3]));
		};
	};

	$_[0]->{triples}->{$t->hashCode()} = $t; #STORE

	$_[0]->{_findIndex}->addLookup($t)
		if(defined $_[0]->{_findIndex});

	$_[0]->updateDigest($t);
};

sub updateDigest {
	croak "Statement ".$_[1]." is not instance of RDF::Statement"
                unless( (defined $_[1]) && ($_[1]->isa('RDF::Statement')));

      	return
		unless(defined $_[0]->{digest});

	# see http://nestroy.wi-inf.uni-essen.de/rdf/sum_rdf_api/#K31
	my $digest = $_[1]->getDigest();

      	RDF::Digest::Util::xor($_[0]->{digest}->getDigestBytes(),$digest->getDigestBytes());
};

sub validLookup {
	return 0;
};

# Removes the triple from the model
sub remove {
	croak "Statement ".$_[1]." is not instance of RDF::Statement"
                unless( (defined $_[1]) && ($_[1]->isa('RDF::Statement')));

	delete $_[0]->{triples}->{$_[1]->hashCode()}; #DELETE

	$_[0]->{_findIndex}->removeLookup($_[1])
		if(defined $_[0]->{_findIndex});

	$_[0]->updateDigest($_[1]);
};

sub isMutable {
	return 1;
};

# General method to search for triples.
# null input for any parameter will match anything.
# Example: Model result = m.find( null, RDF.type, new Resource("http://...#MyClass") )
# finds all instances of the class MyClass
sub find {
	croak "Subject ".$_[1]." is not instance of RDF::Resource"
                unless( (not(defined $_[1])) || ($_[1]->isa('RDF::Resource')));
	croak "Predicate ".$_[2]." is not instance of RDF::Resource"
                unless( (not(defined $_[2])) || ($_[2]->isa('RDF::Resource')));
	croak "Object ".$_[3]." is not instance of RDF::RDFNode"
                unless( (not(defined $_[3])) || ($_[3]->isa('RDF::RDFNode')));

	# we have the same problem like in Pen - a result set must be a model/collection :-)
	#my $res = $_[0]->create(); #EMPTY MODEL
	my $res = new RDF::Store::SetModel(undef,undef,undef,undef);

        return $res
		if($_[0]->isEmpty());

        return $_[0]->duplicate()
                if( (not(defined $_[1])) && (not(defined $_[2])) && (not(defined $_[3])) );

	my $en = (defined $_[0]->{_findIndex}) ? 
			$_[0]->{_findIndex}->multiget($_[1],$_[2],$_[3]) : 
			$_[0]->{triples};

	my $key;
	my $t;
	while(($key,$t) = each %{$en}) {
		my $matchStatement=1;
		if(	( (defined $_[1]) && (!($t->subject()->equals($_[1]))) ) ||
			( (defined $_[2]) && (!($t->predicate()->equals($_[2]))) ) ||
			( (defined $_[3]) && (!($t->object()->equals($_[3]))) ) ) {
			$matchStatement=0;
		};

                if($matchStatement) {
                	$res->add($t);
                };
        };
        return $res;
};

# Clone the model
sub duplicate {
	# creates a model that shares triples and lookup with this model
	# ...I would say definitively PENish :)
	return new RDF::Store::Model($_[0]->{uri}, $_[0]->{triples},
			(defined $_[0]->{_findIndex}) ? $_[0]->{_findIndex} : undef, 1);
};

sub clone {
	return $_[0]->duplicate();
};

# Creates empty model of the same Class
sub create {
	return new RDF::Store::Model(undef,undef,undef,undef);
};

sub getNodeFactory {
	return $_[0]->{nodeFactory};	
};

sub getDigest {
	# we do NOT carry out the whole model digest here - JAVA uses Digestable inteface :(
	if(not(defined $_[0]->{digest})) {
		my $digest_bytes;
		# this does not work under MagicTie yet :(
		my $key;
		my $t;
		while(($key,$t) = each %{$_[0]->{triples}}) {
      			my $d = $t->getDigest();
      			if(not(defined $digest_bytes)) {
        			$digest_bytes = $d->getDigestBytes();
      			} else {
				RDF::Digest::Util::xor($digest_bytes,$d->getDigestBytes());
    			};
		};
		$_[0]->{digest} = RDF::Digest::Util::createFromBytes($_[0]->{algorithm},$digest_bytes);
	};
	return $_[0]->{digest};
};

sub getURI {
	if($_[0]->isEmpty()) {
      		return $_[0]->{nodeFactory}->createUniqueResource()->toString();
    	} else {
		return "uuid:rdf:".
				$_[0]->getDigestAlgorithm ."-".
				RDF::Digest::Util::toHexString($_[0]->getDigest());
      	};
};

sub getDigestAlgorithm {
	return $_[0]->{algorithm};
};

sub getDigestBytes {
	#do something hereish......	
	if(not(defined $_[0]->{digest})) {
		$_[0]->{digest} = $_[0]->getDigest();
	}; 
	#$_[0]->{digest_bytes} = $_[0]->{digest}->getDigestBytes();
        #return \$_[0]->{digest_bytes};
	return $_[0]->{digest}->getDigestBytes();
};

1;
};

__END__

=head1 NAME

RDF::Store::Model - An implementation of the Model RDF API

=head1 SYNOPSIS

	use RDF::Store::Model;
	use RDF::Store::NodeFactory;
	use RDF::Store::FindIndex;
	use Data::MagicTie;

	my $factory= new RDF::Store::NodeFactory();
	my $statement = $factory->createStatement(
                        	$factory->createResource('http://perl.org'),
                        	$factory->createResource('http://iscool.org/schema/1.0/#label'),
                        	$factory->createLiteral('Cool Web site')
                                );
	my $statement1 = $factory->createStatement(
				$factory->createResource("http://www.altavista.com"),
				$factory->createResource("http://pen.jrc.it/schema/1.0/#author"),
				$factory->createLiteral("Who? :-)")
				);

	my $statement2 = $factory->createStatement(
				$factory->createUniqueResource(),
				$factory->createUniqueResource(),
				$factory->createLiteral("")
				);

	my $index_db={};
	tie %{$index_db},"Data::MagicTie",'index/triples',(Q => 20);
	my $index=new RDF::Store::FindIndex($index_db);
	my $model = new RDF::Store::Model($factory,undef,$index,undef);

	$model->add($statement);
	$model->add($statement1);
	$model->add($statement2);
	my $model1 = $model->duplicate();

	print $model1->getDigest->equals( $model1->getDigest );
	print $model1->getDigest->hashCode;

	my $found = $model->find($statement2->subject,undef,undef);
	foreach (keys %{$found->elements}) {
        	print $found->elements->{$_}->getLabel(),"\n";
	};

=head1 DESCRIPTION

An RDF::Model implementation using Digested URIs and Perl hashes to store triplets. The actual store could be tied to a Data::MagicTie(3) hash/array and the RDF::Store::FindIndex(3) module.

=head1 SEE ALSO

Data::MagicTie(3) Digest(3) RDF::Digest(3) RDF::Store::RDFNode RDF::Store::Resource RDF::Store::FindIndex(3)

=head1 AUTHOR

	Alberto Reggiori <alberto.reggiori@jrc.it>
